from huey import RedisHuey
from huey.contrib.helpers import RedisSemaphore
from huey.contrib.helpers import lock_task_semaphore
from huey.exceptions import TaskLockedException
from huey.tests.base import BaseTestCase


class TestLockTaskSemaphore(BaseTestCase):
    def setUp(self):
        super(TestLockTaskSemaphore, self).setUp()
        self.semaphore = RedisSemaphore(self.huey, 'lock_a', 2)
        self.huey.storage.conn.delete(self.semaphore.key)

    def tearDown(self):
        self.huey.storage.conn.delete(self.semaphore.key)
        super(TestLockTaskSemaphore, self).tearDown()

    def get_huey(self):
        return RedisHuey()

    def test_redis_semaphore(self):
        s = self.semaphore
        aid1 = s.acquire()
        self.assertTrue(aid1 is not None)
        aid2 = s.acquire()
        self.assertTrue(aid2 is not None)  # We can acquire it twice.
        self.assertTrue(s.acquire() is None)  # Cannot acquire 3 times.
        self.assertEqual(s.release(aid2), 1)  # Release succeeded.
        self.assertEqual(s.release(aid2), 0)  # Already released.
        self.assertEqual(s.acquire(aid2), aid2)  # Re-acquired.
        self.assertEqual(s.acquire(aid2), aid2)  # No-op (still acquired).

        self.assertEqual(s.release(aid2), 1)  # Release succeeded.
        self.assertEqual(s.release(aid1), 1)  # Release succeeded.

        self.assertTrue(s.acquire() is not None)  # Acquire twice.
        self.assertTrue(s.acquire() is not None)
        self.assertTrue(s.acquire() is None)  # Cannot acquire 3 times.
        self.huey.storage.conn.delete(s.key)



import itertools


SIGNAL_CANCELED = 'canceled'
SIGNAL_COMPLETE = 'complete'
SIGNAL_ERROR = 'error'
SIGNAL_EXECUTING = 'executing'
SIGNAL_EXPIRED = 'expired'
SIGNAL_LOCKED = 'locked'
SIGNAL_RETRYING = 'retrying'
SIGNAL_REVOKED = 'revoked'
SIGNAL_SCHEDULED = 'scheduled'
SIGNAL_INTERRUPTED = 'interrupted'
SIGNAL_ENQUEUED = 'enqueued'


class Signal(object):
    __slots__ = ('receivers',)

    def __init__(self):
        self.receivers = {'any': []}

    def connect(self, receiver, *signals):
        if not signals:
            signals = ('any',)
        for signal in signals:
            self.receivers.setdefault(signal, [])
            self.receivers[signal].append(receiver)

    def disconnect(self, receiver, *signals):
        if not signals:
            signals = list(self.receivers)
        for signal in signals:
            try:
                self.receivers[signal].remove(receiver)
            except ValueError:
                pass

    def send(self, signal, task, *args, **kwargs):
        receivers = itertools.chain(self.receivers.get(signal, ()),
                                    self.receivers['any'])
        for receiver in receivers:
            receiver(signal, task, *args, **kwargs)



import asyncio

from huey.constants import EmptyData


async def aget_result(res, backoff=1.15, max_delay=1.0, preserve=False):
    """
    Await a task result.

    Example usage:

        @huey.task()
        def sleep(n):
            time.sleep(n)
            return n

        # Call the task and get the normal result-handle.
        rh = sleep(2)

        # Asynchronously await the result of the task.
        result = await aget_result(rh)

    More advanced example of waiting for multiple results concurrently:

        r1 = sleep(1)
        r2 = sleep(2)
        r3 = sleep(3)

        # Asynchronously await the results of all 3 tasks. Will take
        # ~3 seconds.
        results = await asyncio.gather(
            aget_result(r1),
            aget_result(r2),
            aget_result(r3))

    NOTE: the Redis operation will be a normal blocking socket read, but in
    practice these will be super fast. The slow part is the necessity to wait
    between polling intervals (since the Redis command to read the result does
    not block).
    """
    delay = 0.1
    while res._result is EmptyData:
        delay = min(delay, max_delay)
        if res._get(preserve) is EmptyData:
            await asyncio.sleep(delay)
            delay *= backoff
    return res._result


async def aget_result_group(rg, *args, **kwargs):
    """
    Await the results of a ResultGroup.

    Example usage:

        @huey.task()
        def sleep(n):
            time.sleep(n)
            return n

        rg = sleep.map([2, 2, 2])

        # This should take ~2 seconds.
        results = await aget_result_group(rg)
    """
    return await asyncio.gather(*[
        aget_result(r, *args, **kwargs)
        for r in rg])



import logging
import optparse
from collections import namedtuple
from logging import FileHandler

from huey.constants import WORKER_THREAD
from huey.constants import WORKER_TYPES


config_defaults = (
    ('workers', 1),
    ('worker_type', WORKER_THREAD),
    ('initial_delay', 0.1),
    ('backoff', 1.15),
    ('max_delay', 10.0),
    ('check_worker_health', True),
    ('health_check_interval', 10),
    ('scheduler_interval', 1),
    ('periodic', True),
    ('logfile', None),
    ('verbose', None),
    ('simple_log', None),
    ('flush_locks', False),
    ('extra_locks', None),
)
config_keys = [param for param, _ in config_defaults]


def option(name, **options):
    if isinstance(name, tuple):
        letter, opt_name = name
    else:
        opt_name = name.replace('_', '-')
        letter = name[0]
        options.setdefault('dest', name)
    return ('-' + letter, '--' + opt_name, options)


class OptionParserHandler(object):
    def get_worker_options(self):
        return (
            # -w, -k, -d, -m, -b, -c, -C, -f
            option('workers', type='int',
                   help='number of worker threads/processes (default=1)'),
            option(('k', 'worker-type'), choices=WORKER_TYPES,
                   dest='worker_type',
                   help=('worker execution model (thread, greenlet, '
                         'process). Use process for CPU-intensive workloads, '
                         'and greenlet for IO-heavy workloads. When in doubt, '
                         'thread is the safest choice.')),
            option('delay', dest='initial_delay',
                   help='minimum time to wait when polling queue (default=.1)',
                   metavar='SECONDS', type='float'),
            option('max_delay', metavar='SECONDS',
                   help='maximum time to wait when polling queue (default=10)',
                   type='float'),
            option('backoff', metavar='SECONDS',
                   help=('factor used to back-off polling interval when queue '
                         'is empty (default=1.15, must be >= 1)'),
                   type='float'),
            option(('c', 'health-check-interval'), type='float',
                   dest='health_check_interval', metavar='SECONDS',
                   help=('minimum time to wait between worker health checks '
                         '(default=1.0)')),
            option(('C', 'disable-health-check'), action='store_false',
                   dest='check_worker_health',
                   help=('disable health check that monitors worker health, '
                         'restarting any worker that crashes unexpectedly.')),
            option('flush_locks', action='store_true', dest='flush_locks',
                   help=('flush all locks when starting consumer.')),
            option(('L', 'extra-locks'), dest='extra_locks',
                   help=('additional locks to flush, separated by comma.')),
        )

    def get_scheduler_options(self):
        return (
            # -s, -n
            option('scheduler_interval', type='int',
                   help='Granularity of scheduler in seconds.'),
            option('no_periodic', action='store_false',
                   dest='periodic', help='do NOT enqueue periodic tasks'),
        )

    def get_logging_options(self):
        return (
            # -l, -v, -q, -S
            option('logfile', metavar='FILE'),
            option('verbose', action='store_true',
                   help='verbose logging (includes DEBUG statements)'),
            option('quiet', action='store_false', dest='verbose',
                   help='minimal logging'),
            option(('S', 'simple'), action='store_true', dest='simple_log',
                   help='simple logging format (time message)'),
        )

    def get_option_parser(self):
        parser = optparse.OptionParser('Usage: %prog [options] '
                                       'path.to.huey_instance')

        def add_group(name, description, options):
            group = parser.add_option_group(name, description)
            for abbrev, name, kwargs in options:
                group.add_option(abbrev, name, **kwargs)

        add_group('Logging', 'The following options pertain to logging.',
                  self.get_logging_options())

        add_group('Workers', (
            'By default huey uses a single worker thread. To specify a '
            'different number of workers, or a different execution model (such'
            ' as multiple processes or greenlets), use the options below.'),
            self.get_worker_options())

        add_group('Scheduler', (
            'By default Huey will run the scheduler once every second to check'
            ' for tasks scheduled in the future, or tasks set to run at '
            'specfic intervals (periodic tasks). Use the options below to '
            'configure the scheduler or to disable periodic task scheduling.'),
            self.get_scheduler_options())

        return parser


class ConsumerConfig(namedtuple('_ConsumerConfig', config_keys)):
    def __new__(cls, **kwargs):
        config = dict(config_defaults)
        config.update(kwargs)
        args = [config[key] for key in config_keys]
        return super(ConsumerConfig, cls).__new__(cls, *args)

    def validate(self):
        if self.backoff < 1:
            raise ValueError('The backoff must be greater than 1.')
        if not (0 < self.scheduler_interval <= 60):
            raise ValueError('The scheduler must run at least once per '
                             'minute, and at most once per second (1-60).')
        if 60 % self.scheduler_interval != 0:
            raise ValueError('The scheduler interval must be a factor of 60: '
                             '1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, or 60')

    @property
    def loglevel(self):
        if self.verbose is None:
            return logging.INFO
        return logging.DEBUG if self.verbose else logging.WARNING

    def setup_logger(self, logger=None):
        if self.worker_type == 'process':
            worker = '%(process)d'
        else:
            worker = '%(threadName)s'

        if self.simple_log:
            datefmt = '%H:%M:%S'
            logformat = '%(asctime)s %(message)s'
        else:
            datefmt = None  # Use default
            logformat = ('[%(asctime)s] %(levelname)s:%(name)s:' + worker +
                         ':%(message)s')
        if logger is None:
            logger = logging.getLogger()

        if self.logfile:
            handler = logging.FileHandler(self.logfile)
        else:
            handler = logging.StreamHandler()

        handler.setFormatter(logging.Formatter(logformat, datefmt))
        logger.addHandler(handler)
        logger.setLevel(self.loglevel)

    @property
    def values(self):
        return dict((key, getattr(self, key)) for key in config_keys
                    if key not in ('logfile', 'verbose', 'simple_log'))
